searchData={"items":[{"type":"module","title":"AbsintheRelayKeysetConnection","doc":"Support for paginated result sets using keyset pagination, for use in an\nAbsinthe resolver module.\nRequires defining a connection with\n[Absinthe.Relay.Connection](https://hexdocs.pm/absinthe_relay/Absinthe.Relay.Connection.html).","ref":"AbsintheRelayKeysetConnection.html"},{"type":"module","title":"The TL;DR - AbsintheRelayKeysetConnection","doc":"In a resolver...\n\n```elixir\nAbsintheRelayKeysetConnection.from_query(\n  ecto_users_query,\n  &MyRepo.all/1,\n  %{\n    sorts: [%{name: :desc}, %{id: :desc}],\n    first: 10,\n    after: \"0QTwn5SRWyJNbyIsMjZd\"\n  },\n  %{\n    unique_column: :id\n  }\n)\n```","ref":"AbsintheRelayKeysetConnection.html#module-the-tl-dr"},{"type":"module","title":"Details - AbsintheRelayKeysetConnection","doc":"","ref":"AbsintheRelayKeysetConnection.html#module-details"},{"type":"module","title":"Why keyset pagination? - AbsintheRelayKeysetConnection","doc":"By default,\n[Absinthe.Relay.Connection](https://hexdocs.pm/absinthe_relay/Absinthe.Relay.Connection.html)\nuses offset-based pagination.\nFor example, with a page size of 10, it would get the first page of records\nwith a SQL query like `OFFSET 0 LIMIT 10`, the second page with `OFFSET 10\nLIMIT 10`, and so on.\nThis works well for many use cases and requires no knowledge of the\nunderlying database schema.\nHowever, when the value of `OFFSET` is large, it [can cause poor database\nperformance](https://use-the-index-luke.com/no-offset).\n\n(A \"large\" offset depends on your data; it might be in the millions before you\nnotice a performance hit.)\n\nKeyset pagination means that, in the example above, the first page might be\nfetched with `WHERE id > 0 ORDER BY id ASC LIMIT 10`.\nIf the last record on that page had id `10`, the query for the next page could be\nfetched with `WHERE id > 10 ORDER BY id ASC LIMIT 10`.\nThis `WHERE` clause lets the database efficiently ignore earlier records,\nespecially if the `id` column is indexed.","ref":"AbsintheRelayKeysetConnection.html#module-why-keyset-pagination"},{"type":"module","title":"The cursor - AbsintheRelayKeysetConnection","doc":"With offset-based pagination, a user needs only to say \"I want 10 records per\npage, and give me page 3, please.\"\nWe can easily calculate the offset as `(page_number - 1) * limit`.\n\nBut for keyset-based pagination, we need more information.\nTo get the next page, we need to know which record appeared last on the page\nthe user just got; for example, if it was record 10, we will query `WHERE id\n> 10`.\nThe user needs to supply this information using a \"cursor\".\nIn this simple case, the cursor need contain only the id.\n(Typically this value is encoded in a way that makes it opaque to the user in\norder to indicate that it's an implementation detail.)\nBut sorting and the need for uniqueness add some complexity to the picture.","ref":"AbsintheRelayKeysetConnection.html#module-the-cursor"},{"type":"module","title":"Sorting and uniqueness - AbsintheRelayKeysetConnection","doc":"Keyset pagination only works if our sorting (eg `ORDER BY id asc` and\ncomparison (eg `WHERE id > 10`) agree and are based on a unique, sortable\ncolumn or combination of columns.\n(In database terms, any attribute or list of attributes which uniquely identify\na row is a \"key\"; this likely is the basis of the term \"keyset pagination\".)\n\nImagine trying to use a non-unique column like `last_name`.\nIf the last person on the current page is `Abe Able`, requesting the next\npage with `WHERE last_name > 'Able'` will accidentally skip `Beth Able`, who should\nhave appeared on the next page.\n\nTo avoid this, we need to ensure that we order by a unique combination of\ncolumns - such as `ORDER BY last_name ASC, id ASC` - and use the same columns for the\n`WHERE` - such as `WHERE last_name > 'Able' OR (last_name = 'Able' AND id >\n  10)`.\n\nIf your table has a unique column like `id`, `from_query/4` can automatically\nadd it to the `ORDER BY` and `WHERE` clauses of queries which don't already\nuse it; just indicate which column to use in the `config` argument.\n\nIt's also problematic if the columns are not meaningfully ordered. If you're\npaging through records and request `WHERE id > 100`, a new record inserted\nwith id `105` will show up in your results.\nBut if you're using `WHERE uuid > $some_uuid`, it might not.\n\nNote that since the cursor is the basis of the `WHERE` clause, whatever\ncolumns the query is being ordered by are included in the cursor value (which\nis opaque to users).\nIn the example above, the cursor for each record would include the last name\nand id.","ref":"AbsintheRelayKeysetConnection.html#module-sorting-and-uniqueness"},{"type":"module","title":"Serialization - AbsintheRelayKeysetConnection","doc":"As explained above, building the cursor involves serializing the columns\nwhich are used in the `ORDER BY` so that they can also be used in the\n`WHERE`.\nFor example, if ordering users by name and id, the cursor for each user\nrecord will contain that user's name and id.\n\nYou can provide your own `AbsintheRelayKeysetConnection.CursorTranslator` to customize\nthe serialization and deserialization of your cursors, or use the built-in\n`AbsintheRelayKeysetConnection.CursorTranslator.Base64Hashed`.","ref":"AbsintheRelayKeysetConnection.html#module-serialization"},{"type":"module","title":"Handling NULL Values - AbsintheRelayKeysetConnection","doc":"By default, NULL values in sortable columns can cause pagination to fail because\nSQL comparisons with NULL (like `WHERE name > NULL`) are unsafe and forbidden by Ecto.\n\nTo handle this, you can use the `:null_coalesce` configuration option to specify\nreplacement values for NULL columns during sorting and cursor operations.","ref":"AbsintheRelayKeysetConnection.html#module-handling-null-values"},{"type":"module","title":"Example with NULL Coalescing - AbsintheRelayKeysetConnection","doc":"```elixir\n# Without null coalescing - crashes if cursors contain NULL values\nAbsintheRelayKeysetConnection.from_query(\n  User,\n  &MyRepo.all/1,\n  %{sorts: [%{last_name: :asc}], first: 10},\n  %{unique_column: :id}\n)\n\n# With null coalescing - NULL values are treated as empty strings\nAbsintheRelayKeysetConnection.from_query(\n  User,\n  &MyRepo.all/1,\n  %{sorts: [%{last_name: :asc}], first: 10},\n  %{\n    unique_column: :id,\n    null_coalesce: %{last_name: \"\"}\n  }\n)\n```","ref":"AbsintheRelayKeysetConnection.html#module-example-with-null-coalescing"},{"type":"module","title":"How NULL Coalescing Works - AbsintheRelayKeysetConnection","doc":"When you configure `null_coalesce: %{last_name: \"\"}`:\n\n1. **SQL Sorting**: Uses `ORDER BY COALESCE(last_name, '')` - NULL values sort as empty strings\n2. **Cursor Encoding**: NULL values are encoded as the coalesce value in cursors\n3. **WHERE Clauses**: Uses `WHERE COALESCE(last_name, '') > 'value'` for safe comparisons","ref":"AbsintheRelayKeysetConnection.html#module-how-null-coalescing-works"},{"type":"module","title":"Per-Column Configuration - AbsintheRelayKeysetConnection","doc":"You can specify different coalesce values for different columns:\n\n```elixir\n%{\n  null_coalesce: %{\n    last_name: \"\",           # NULL last names become empty string\n    first_name: \"Unknown\",   # NULL first names become \"Unknown\"\n    score: 0                 # NULL scores become 0\n  }\n}\n```\n\nThis ensures consistent, predictable sorting behavior regardless of NULL values in your data.","ref":"AbsintheRelayKeysetConnection.html#module-per-column-configuration"},{"type":"module","title":"Cautions - AbsintheRelayKeysetConnection","doc":"There are a few things you can't do with this pagination style.\n\nFirst, you can't (reliably) paginate records without specifying a unique\ncolumn or combination of columns.\nFor example, if you have an `cities` table where the primary key is the\ncombination of `state_id` and `city_name`, you'll have to ensure that all\nqueries include both values in their `sorts`.\n(It might be simpler to add a sequential integer column and pass that as the\n`:unique_column`.)\n\nSecond, you can't sort by columns on associations.\nThis is not implemented and would be difficult to implement.\n\nThird, you can't paginate at more than one level.\nFor example, you can't get the first page of authors, get the first page of\nposts for each author, and proceed to get subsequent pages of posts for each\nauthor.\nSuch an access pattern is not a good idea even with `OFFSET` pagination; some\nauthors will have many more pages of posts than others.\nBut it becomes truly nonsensical to say \"for each author, get the first 10\nposts with id greater than 10\".\n\nInstead of attempting this, it would be better to paginate authors, then\nseparately, paginate posts filtered by author id.\n\nFinally, a caution: since the cursor needs to contain the column values,\nbeware of using large columns in a cursor; they will make the cursor itself\nlarge.","ref":"AbsintheRelayKeysetConnection.html#module-cautions"},{"type":"type","title":"AbsintheRelayKeysetConnection.column_name/0","doc":"The name of a column to be used in an `ORDER BY` clause.","ref":"AbsintheRelayKeysetConnection.html#t:column_name/0"},{"type":"type","title":"AbsintheRelayKeysetConnection.config/0","doc":"Options that are independent of the current query document.","ref":"AbsintheRelayKeysetConnection.html#t:config/0"},{"type":"type","title":"AbsintheRelayKeysetConnection.edge/0","doc":"A wrapper for a single record which includes the record itself (the node) and\na cursor that references it.","ref":"AbsintheRelayKeysetConnection.html#t:edge/0"},{"type":"type","title":"AbsintheRelayKeysetConnection.edge_node/0","doc":"A single record.","ref":"AbsintheRelayKeysetConnection.html#t:edge_node/0"},{"type":"type","title":"AbsintheRelayKeysetConnection.encoded_cursor/0","doc":"A pagination cursor which is encoded and opaque to users. A cursor represents\nthe position of a specific record in the pagination set. For example, the\ncursor given with post `20` represents that post, so that a user can make a\nfollow-up request using the same `sorts` but specifying the first 10 records\nafter post `20`, the last 5 records before post `20`, or something similar.","ref":"AbsintheRelayKeysetConnection.html#t:encoded_cursor/0"},{"type":"function","title":"AbsintheRelayKeysetConnection.from_query/4","doc":"Build a connection from an Ecto Query.\n\nThis will automatically set an `ORDER BY` and `WHERE` value based on the\nprovided options, including the cursor (if one is given), then run the query\nwith the `repo_fun` argument that was given.\n\nReturn a single page of results which contains the info specified in the\n[Relay Cursor Connections\nSpecification](https://relay.dev/graphql/connections.htm).","ref":"AbsintheRelayKeysetConnection.html#from_query/4"},{"type":"function","title":"Example - AbsintheRelayKeysetConnection.from_query/4","doc":"iex> AbsintheRelayKeysetConnection.from_query(\n    ...>   ecto_users_query,\n    ...>   &MyRepo.all/1,\n    ...>   %{\n    ...>     sorts: [%{name: :desc}, %{id: :desc}],\n    ...>     first: 10,\n    ...>     after: \"0QTwn5SRWyJNbyIsMjZd\"\n    ...>   },\n    ...>   %{\n    ...>     unique_column: :id,\n    ...>     null_coalesce: %{name: \"\"}\n    ...>   }\n    ...> )\n    {:ok, %{\n      edges: [\n        %{node: %MyApp.User{id: 11, name: \"Jo\"}, cursor: \"abc123\"},\n        %{node: %MyApp.User{id: 12, name: \"Mo\"}, cursor: \"def345\"}\n      ],\n      page_info: %{\n        start_cursor: \"abc123\",\n        end_cursor: \"def345\",\n        has_previous_page: true,\n        has_next_page: false\n      }\n    }}","ref":"AbsintheRelayKeysetConnection.html#from_query/4-example"},{"type":"type","title":"AbsintheRelayKeysetConnection.options/0","doc":"Options derived from the current query document.","ref":"AbsintheRelayKeysetConnection.html#t:options/0"},{"type":"type","title":"AbsintheRelayKeysetConnection.page_info/0","doc":"Information about the set of records in the current page and how it relates\nto the overall set of records available for pagination.","ref":"AbsintheRelayKeysetConnection.html#t:page_info/0"},{"type":"type","title":"AbsintheRelayKeysetConnection.repo_fun/0","doc":"A function which can take an `Ecto.Queryable()` and use it to fetch records\nfrom a data store.\nA common example would be `&MyRepo.all/1`.","ref":"AbsintheRelayKeysetConnection.html#t:repo_fun/0"},{"type":"type","title":"AbsintheRelayKeysetConnection.sort/0","doc":"A single-key map, such as `%{name: :asc}`.\n\nThis is the information needed to build a single `ORDER BY` clause.","ref":"AbsintheRelayKeysetConnection.html#t:sort/0"},{"type":"type","title":"AbsintheRelayKeysetConnection.sort_dir/0","doc":"Either `:asc` or `:desc`, to be used in an `ORDER BY` clause.","ref":"AbsintheRelayKeysetConnection.html#t:sort_dir/0"},{"type":"type","title":"AbsintheRelayKeysetConnection.t/0","doc":"The return value of `from_query/4`, representing the paginated data.","ref":"AbsintheRelayKeysetConnection.html#t:t/0"},{"type":"behaviour","title":"AbsintheRelayKeysetConnection.CursorTranslator","doc":"A cursor translator handles encoding and decoding pagination cursors.\n\nThis module provides the behaviour for implementing a cursor translator.\nAn example use case of this module would be a cursor translator that\nencodes and decodes signed or encrypted pagination cursors.","ref":"AbsintheRelayKeysetConnection.CursorTranslator.html"},{"type":"behaviour","title":"Example Cursor Translator - AbsintheRelayKeysetConnection.CursorTranslator","doc":"A basic example that encodes and decodes cursors using Jason and base64:\n\n```elixir\ndefmodule MyApp.Absinthe.Cursor do\n  @behaviour AbsintheRelayKeysetConnection.CursorTranslator\n\n  @impl AbsintheRelayKeysetConnection.CursorTranslator\n  def from_key(key_map, cursor_columns, config \\ %{}) do\n    null_coalesce = Map.get(config, :null_coalesce, %{})\n    \n    values = \n      cursor_columns\n      |> Enum.map(fn column ->\n        value = Map.fetch!(key_map, column)\n        case value do\n          nil -> Map.get(null_coalesce, column, nil)\n          _ -> value\n        end\n      end)\n    \n    values\n    |> Jason.encode!()\n    |> Base.encode64(padding: false)\n  end\n\n  @impl AbsintheRelayKeysetConnection.CursorTranslator\n  def to_key(encoded_cursor, expected_columns, _config \\ %{}) do\n    values = encoded_cursor |> Base.decode64!(padding: false) |> Jason.decode!()\n    key = expected_columns |> Enum.zip(values) |> Map.new()\n\n    {:ok, key}\n  end\nend\n```","ref":"AbsintheRelayKeysetConnection.CursorTranslator.html#module-example-cursor-translator"},{"type":"type","title":"AbsintheRelayKeysetConnection.CursorTranslator.column_key/0","doc":"A key of the column of the node being paginated.","ref":"AbsintheRelayKeysetConnection.CursorTranslator.html#t:column_key/0"},{"type":"type","title":"AbsintheRelayKeysetConnection.CursorTranslator.encoded_cursor/0","doc":"A string containing the encoded cursor.","ref":"AbsintheRelayKeysetConnection.CursorTranslator.html#t:encoded_cursor/0"},{"type":"callback","title":"AbsintheRelayKeysetConnection.CursorTranslator.from_key/2","doc":"Creates the cursor string from a key.\n\nConverts a key map into a cursor string using the given cursor columns.","ref":"AbsintheRelayKeysetConnection.CursorTranslator.html#c:from_key/2"},{"type":"callback","title":"AbsintheRelayKeysetConnection.CursorTranslator.from_key/3","doc":"Creates the cursor string from a key with additional configuration.\n\nConverts a key map into a cursor string using the given cursor columns,\napplying any configured transformations like null coalescing.","ref":"AbsintheRelayKeysetConnection.CursorTranslator.html#c:from_key/3"},{"type":"callback","title":"Configuration Options - AbsintheRelayKeysetConnection.CursorTranslator.from_key/3","doc":"- `:null_coalesce` - Map of column names to coalesce values for NULL handling","ref":"AbsintheRelayKeysetConnection.CursorTranslator.html#c:from_key/3-configuration-options"},{"type":"type","title":"AbsintheRelayKeysetConnection.CursorTranslator.key_map/0","doc":"A map containing the node data.","ref":"AbsintheRelayKeysetConnection.CursorTranslator.html#t:key_map/0"},{"type":"callback","title":"AbsintheRelayKeysetConnection.CursorTranslator.to_key/2","doc":"Rederives the key from the cursor string.\n\nConverts a cursor string into a key map using the given cursor columns.","ref":"AbsintheRelayKeysetConnection.CursorTranslator.html#c:to_key/2"},{"type":"callback","title":"AbsintheRelayKeysetConnection.CursorTranslator.to_key/3","doc":"Rederives the key from the cursor string with additional configuration.\n\nConverts a cursor string into a key map using the given cursor columns,\napplying any configured transformations like null coalescing.","ref":"AbsintheRelayKeysetConnection.CursorTranslator.html#c:to_key/3"},{"type":"callback","title":"Configuration Options - AbsintheRelayKeysetConnection.CursorTranslator.to_key/3","doc":"- `:null_coalesce` - Map of column names to coalesce values for NULL handling","ref":"AbsintheRelayKeysetConnection.CursorTranslator.html#c:to_key/3-configuration-options"},{"type":"module","title":"AbsintheRelayKeysetConnection.CursorTranslator.Base64Hashed","doc":"A cursor translator implementation that uses base64 and a hashed padding.\n\nA tamper-resistant (not tamper-proof) implementation that uses base64 and a hashed padding.\n\nThese values are serialized using `Jason.encode/1`, which means you'll need\nan implementation of the `Jason.Encoder` protocol for the type of each column you\nsort by.\nThe library covers most common data types, but you may need to implement your\nown for less common ones.\n\nFor example, if you're using `Postgrex.INET` for a PostgreSQL `inet` column,\nyou might need:\n\n```elixir\ndefmodule MyApp.CustomEncoders do\n defimpl Jason.Encoder, for: [Postgrex.INET] do\n   def encode(struct, opts) do\n     Jason.Encode.string(EctoNetwork.INET.decode(struct), opts)\n   end\n end\nend","ref":"AbsintheRelayKeysetConnection.CursorTranslator.Base64Hashed.html"},{"type":"function","title":"AbsintheRelayKeysetConnection.CursorTranslator.Base64Hashed.from_key/3","doc":"Creates the cursor string from a key.\nThis encoding is not meant to be tamper-proof, just to hide the cursor data\nas an implementation detail.","ref":"AbsintheRelayKeysetConnection.CursorTranslator.Base64Hashed.html#from_key/3"},{"type":"function","title":"Examples - AbsintheRelayKeysetConnection.CursorTranslator.Base64Hashed.from_key/3","doc":"iex> from_key(%{id: 25}, [:id])\n    \"Tr7wn5SRWzI1XQ==\"\n\n    iex> from_key(%{name: \"Mo\", id: 26}, [:name, :id])\n    \"eo7wn5SRWyJNbyIsMjZd\"","ref":"AbsintheRelayKeysetConnection.CursorTranslator.Base64Hashed.html#from_key/3-examples"},{"type":"function","title":"AbsintheRelayKeysetConnection.CursorTranslator.Base64Hashed.to_key/3","doc":"Rederives the key from the cursor string.\nThe cursor string is supplied by users and may have been tampered with.\nHowever, we ensure that only the expected column values may appear in the\ncursor, so at worst, they could paginate from a different spot, which is\nfine.","ref":"AbsintheRelayKeysetConnection.CursorTranslator.Base64Hashed.html#to_key/3"},{"type":"function","title":"Examples - AbsintheRelayKeysetConnection.CursorTranslator.Base64Hashed.to_key/3","doc":"iex> to_key(\"Tr7wn5SRWzI1XQ==\", [:id])\n    {:ok, %{id: 25}}","ref":"AbsintheRelayKeysetConnection.CursorTranslator.Base64Hashed.html#to_key/3-examples"},{"type":"extras","title":"AbsintheRelayKeysetConnection","doc":"# AbsintheRelayKeysetConnection\n\n[![Version](https://img.shields.io/hexpm/v/absinthe_relay_keyset_connection.svg)](https://hex.pm/packages/absinthe_relay_keyset_connection)\n[![Hex Docs](https://img.shields.io/badge/documentation-gray.svg)](https://hexdocs.pm/absinthe_relay_keyset_connection)\n\nSupport for paginated result sets using keyset pagination, for use in an\nAbsinthe resolver module.\nRequires defining a connection with\n[Absinthe.Relay.Connection](https://hexdocs.pm/absinthe_relay/Absinthe.Relay.Connection.html).","ref":"readme.html"},{"type":"extras","title":"Installation - AbsintheRelayKeysetConnection","doc":"Add to to your list of dependencies in `mix.exs`:\n\n```elixir\ndef deps do\n  [\n    {:absinthe_relay_keyset_connection, \"~> 2.0\"}\n  ]\nend\n```\n\n... and `mix deps.get`.","ref":"readme.html#installation"},{"type":"extras","title":"Usage - AbsintheRelayKeysetConnection","doc":"See the docs linked from [the package](https://hex.pm/packages/absinthe_relay_keyset_connection).","ref":"readme.html#usage"},{"type":"extras","title":"CHANGELOG","doc":"# CHANGELOG","ref":"changelog.html"},{"type":"extras","title":"v2.1.0 (2025-07-10) - CHANGELOG","doc":"- Add `null_coalesce` option, thanks to @tomasz-tomczyk","ref":"changelog.html#v2-1-0-2025-07-10"},{"type":"extras","title":"v2.0.0 (2025-06-21) - CHANGELOG","doc":"- Now requires Elixir v1.15\n- Updated minimum version of dependencies:\n  - ecto_sql ~> 3.8\n  - ex_doc ~> 0.34\n  - dialyxir ~> 1.4","ref":"changelog.html#v2-0-0-2025-06-21"},{"type":"extras","title":"v1.0.0 (2022-04-01) - CHANGELOG","doc":"- First public release","ref":"changelog.html#v1-0-0-2022-04-01"}],"proglang":"elixir","content_type":"text/markdown","producer":{"name":"ex_doc","version":"0.38.2"}}